'''State snapshots to facilitate debugging and fixing apt-listchanges bugs

Sometimes apt-listchanges malfunctions in ways that are highly dependent on the
environment in which it's running and the specific packages it's looking for
changes in. We've implemented this snapshot functionality to facilitate finding
and fixing those bugs quickly when there is active development of the program
ongoing and therefore odds are higher that bugs will be introduced.

When snapshots are enabled (more on that below), each time the program runs, it
makes copies of several significant files and also saves some data to disk, in
a snapshot folder which is, at the end of the run, turned into a .tar.xz file.
The last seven .tar.xz files are saved. If a user who has snapshots enabled
reports a bug, we can ask them to send us the relevant snapshot to assist us in
troubleshooting.

Any user can enable snapshots at any time by adding the configuration settings
`capture_snapshots=true` and `snapshot_dir=[directory-path]` to the
apt-listchanges config fragment file for a particular profile. However, these
settings are undocumented and generally not expected to be added by hand.
Instead, when debconf_helper.py is configured to enable snapshots, it adds
`capture_snapshots=auto` and `snapshot_dir` settings to the configuration of
the `apt` profile.

Setting `capture_snapshots=auto` is equivalent to setting it to `true`, but it
means that in a subsequent release when we change debconf_helper.py to no
longer enable snapshots, it's allowed to automatically remove the
`capture_snapshots` and `snapshot_dir` options and delete any snapshots that
are still around. In this way we can transition easily from experimental to
non-experimental packages and snapshots get disabled and cleaned up
automatically.

The build checks to make sure that debconf_helper.py doesn't have snapshots
enabled unless the changelog says we're doing an experimental release. We never
want to release apt-listchanges past experimental with snapshots enabled.

'''

import datetime
import os
import shutil

from apt_listchanges.ALCConfig import ALCConfig
from apt_listchanges.ALCLog import debug


class Snapshot:
    '''Build a snapshot of what happened during the apt-listchanges run
    This is designed to _never cause an exception_. If something goes wrong it
    just gives up. The goal here is to make a best effort and never interfere
    with the proper operation of the program.'''
    SAVED_SNAPSHOTS = 7

    def __init__(self, config: ALCConfig) -> None:
        self.archive = None
        if not config.capture_snapshots:
            self.capturing = False
            return
        if not config.snapshot_dir:
            self.capturing = False
            return
        self.snapshot_dir = config.snapshot_dir
        if not os.path.exists(self.snapshot_dir):
            if not os.path.exists(os.path.dirname(self.snapshot_dir)):
                # It's too risky to go traipsing around the filesystem creating
                # multiple levels of directories, so let's just skip it. If the
                # user really wants to store snapshots multiple levels deep in
                # a new hierarchy, they can create the directory by hand.
                self.capturing = False
                return
            try:
                os.mkdir(self.snapshot_dir)
            except Exception as ex:
                debug(f'mkdir({self.snapshot_dir}) failed, {ex}')
                self.capturing = False
                return
        self.snapshot_subdir = datetime.datetime.now().replace(microsecond=0).\
            isoformat()
        self.snapshot_path = os.path.join(
            self.snapshot_dir, self.snapshot_subdir)
        try:
            os.mkdir(self.snapshot_path)
        except Exception as ex:
            debug(f'mkdir({self.snapshot_path}) failed, {ex}')
            self.capturing = False
            return
        self.capturing = True

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _exc_traceback):
        self.commit()
        return False

    def abort(self):
        self.capturing = False
        try:
            shutil.rmtree(self.snapshot_path, ignore_errors=True)
        except Exception:  # pragma: no cover
            pass

    def add_file(self, path: str, name: str | None = None) -> None:
        '''Add a file to the snapshot
        If name is not specified the the basename of the file is used'''
        if not self.capturing:
            return
        if not name:
            name = os.path.basename(path)
        if not os.path.exists(path):
            self.add_data('', f'{name}.missing')
            return
        try:
            shutil.copy(path, os.path.join(self.snapshot_path, name))
        except Exception:
            self.abort()

    def add_data(self, data: str | bytes, name: str) -> None:
        if not self.capturing:
            return
        if isinstance(data, str):
            mode = 'wt'
            encoding = {'encoding': 'utf-8'}
        else:
            mode = 'wb'
            encoding = {}
        try:
            target = os.path.join(self.snapshot_path, name)
            # pylint: disable=unspecified-encoding
            with open(target, mode, **encoding) as f:
                f.write(data)
        except Exception:  # pragma: no cover
            self.abort()

    def commit(self):
        if not self.capturing:
            return
        try:
            self.archive = shutil.make_archive(
                self.snapshot_path, 'xztar', self.snapshot_dir,
                self.snapshot_subdir)
        except Exception:
            self.abort()
            return
        self.capturing = False
        try:
            shutil.rmtree(self.snapshot_path, ignore_errors=True)
        except Exception:  # pragma: no cover
            return
        self.prune()

    def prune(self):
        '''Prune old snapshots or snapshot directories'''
        try:
            snapshots = sorted(
                (de.name for de in os.scandir(self.snapshot_dir)),
                reverse=True)[self.SAVED_SNAPSHOTS:]
            for snapshot in snapshots:
                path = os.path.join(self.snapshot_dir, snapshot)
                try:
                    os.remove(path)
                except Exception:  # pragma: no cover
                    shutil.rmtree(path, ignore_errors=True)
        except Exception:  # pragma: no cover
            pass



"""
Apply Debian-specific patches to distutils commands.

Extracts the customized behavior from patches as reported
in pypa/distutils#2 and applies those customizations (except
for scheme definitions) to those commands.

Place this module somewhere in sys.path to take effect.
"""

import os
import sys

import distutils.sysconfig
import distutils.command.install as orig_install
import distutils.command.install_egg_info as orig_install_egg_info
from distutils.command.install_egg_info import (
    to_filename,
    safe_name,
    safe_version,
    )
from distutils.errors import DistutilsOptionError


class install(orig_install.install):
    user_options = list(orig_install.install.user_options) + [
        ('install-layout=', None,
         "installation layout to choose (known values: deb, unix)"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.prefix_option = None
        self.install_layout = None

    def select_scheme(self, name):
        if name == "posix_prefix":
            if self.install_layout:
                if self.install_layout.lower() in ['deb']:
                    name = "deb_system"
                elif self.install_layout.lower() in ['unix']:
                    name = "posix_prefix"
                else:
                    raise DistutilsOptionError(
                        "unknown value for --install-layout")
            elif ((self.prefix_option and
                   os.path.normpath(self.prefix) != '/usr/local')
                  or is_virtual_environment()):
                name = "posix_prefix"
            else:
                if os.path.normpath(self.prefix) == '/usr/local':
                    self.prefix = self.exec_prefix = '/usr'
                    self.install_base = self.install_platbase = '/usr'
                name = "posix_local"
        super().select_scheme(name)

    def finalize_unix(self):
        self.prefix_option = self.prefix
        super().finalize_unix()


class install_egg_info(orig_install_egg_info.install_egg_info):
    user_options = list(orig_install_egg_info.install_egg_info.user_options) + [
        ('install-layout', None, "custom installation layout"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.prefix_option = None
        self.install_layout = None

    def finalize_options(self):
        self.set_undefined_options('install',('install_layout','install_layout'))
        self.set_undefined_options('install',('prefix_option','prefix_option'))
        super().finalize_options()

    @property
    def basename(self):
        if self.install_layout:
            if not self.install_layout.lower() in ['deb', 'unix']:
                raise DistutilsOptionError(
                    "unknown value for --install-layout")
            no_pyver = (self.install_layout.lower() == 'deb')
        elif self.prefix_option:
            no_pyver = False
        else:
            no_pyver = True
        if no_pyver:
            basename = "%s-%s.egg-info" % (
                to_filename(safe_name(self.distribution.get_name())),
                to_filename(safe_version(self.distribution.get_version()))
                )
        else:
            basename = "%s-%s-py%d.%d.egg-info" % (
                to_filename(safe_name(self.distribution.get_name())),
                to_filename(safe_version(self.distribution.get_version())),
                *sys.version_info[:2]
            )
        return basename


def is_virtual_environment():
    return sys.base_prefix != sys.prefix or hasattr(sys, "real_prefix")


def _posix_lib(standard_lib, libpython, early_prefix, prefix):
    is_default_prefix = not early_prefix or os.path.normpath(early_prefix) in ('/usr', '/usr/local')
    if standard_lib:
        return libpython
    elif is_default_prefix and not is_virtual_environment():
        return os.path.join(prefix, "lib", "python3", "dist-packages")
    else:
        return os.path.join(libpython, "site-packages")


def _inject_headers(name, scheme):
    """
    Given a scheme name and the resolved scheme,
    if the scheme does not include headers, resolve
    the fallback scheme for the name and use headers
    from it. pypa/distutils#88

    headers: module headers install location (posix_local is /local/ prefixed)
    include: cpython headers (Python.h)
    See also: bpo-44445
    """
    if 'headers' not in scheme:
        if name == 'posix_prefix':
            headers = scheme['include']
        else:
            headers = orig_install.INSTALL_SCHEMES['posix_prefix']['headers']
        if name == 'posix_local' and '/local/' not in headers:
            headers = headers.replace('/include/', '/local/include/')
        scheme['headers'] = headers
    return scheme


def load_schemes_wrapper(_load_schemes):
    """
    Implement the _inject_headers modification, above, but before
    _inject_headers() was introduced, upstream. So, slower and messier.
    """
    def wrapped_load_schemes():
        schemes = _load_schemes()
        for name, scheme in schemes.items():
            _inject_headers(name, scheme)
        return schemes
    return wrapped_load_schemes


def apply_customizations():
    orig_install.install = install
    orig_install_egg_info.install_egg_info = install_egg_info
    distutils.sysconfig._posix_lib = _posix_lib

    if hasattr(orig_install, '_inject_headers'):
        # setuptools-bundled distutils >= 60.0.5
        orig_install._inject_headers = _inject_headers
    elif hasattr(orig_install, '_load_schemes'):
        # setuptools-bundled distutils >= 59.2.0
        orig_install._load_schemes = load_schemes_wrapper(orig_install._load_schemes)


apply_customizations()



# vim:set fileencoding=utf-8 et ts=4 sts=4 sw=4:
#
#   apt-listchanges - Show changelog entries between the installed versions
#                     of a set of packages and the versions contained in
#                     corresponding .deb files
#
#   Copyright (C) 2016       Robert Luberda  <robert@debian.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

import sys

from apt_listchanges.ALChacks import _

debug_enabled = False


def set_debug(should_debug):
    global debug_enabled  # pylint: disable=global-statement
    debug_enabled = should_debug


def error(msg):
    print(_("apt-listchanges: %(msg)s") % {'msg': msg}, file=sys.stderr)


def warning(msg):
    print(_("apt-listchanges warning: %(msg)s") % {'msg': msg},
          file=sys.stderr)


def info(msg):
    print(_("apt-listchanges: %(msg)s") % {'msg': msg}, file=sys.stdout)


def debug(msg):
    if debug_enabled:
        print(msg, file=sys.stdout)



# Copyright:
#   Moshe Zadka (c) 2002
#   Canonical Ltd. (c) 2005 (DebconfCommunicator)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY AUTHORS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

from __future__ import annotations

import errno
import fcntl
import os
import re
import subprocess
import sys
from types import TracebackType
from typing import IO, Protocol


class Command(Protocol):
    def __call__(self, *params: str | int) -> str: ...


class DebconfError(Exception):
    pass


LOW, MEDIUM, HIGH, CRITICAL = "low", "medium", "high", "critical"


class Debconf:
    """A class that speaks the debconf protocol.

    The simplest way to use this is as a context manager:

        import debconf

        with debconf.Debconf(run_frontend=True) as db:
            print(db.get('debconf/frontend'))

    Note that this will send the STOP command on exiting the context
    manager, so you shouldn't expect to be able to use the same frontend
    again after this.  If you need to do that, then you should instantiate
    the class directly instead:

        import debconf

        db = debconf.Debconf(run_frontend=True)
        print(db.get('debconf/frontend'))
    """

    beginblock: Command
    capb: Command
    endblock: Command
    exist: Command
    fset: Command
    get: Command
    go: Command
    info: Command
    input: Command
    progress: Command
    purge: Command
    register: Command
    reset: Command
    set: Command
    settitle: Command
    subst: Command
    title: Command
    unregister: Command
    version_: Command
    visible: Command

    def __init__(
        self,
        title: str | None = None,
        read: IO[str] | None = None,
        write: IO[str] | None = None,
        run_frontend: bool = False,
    ) -> None:
        for command in (
            "capb set reset title input beginblock endblock go get"
            " register unregister subst fset fget"
            " visible purge metaget exist version_ settitle"
            " info progress data"
        ).split():
            self.setCommand(command)
        self.read = read or sys.stdin
        self.write = write or sys.stdout
        sys.stdout = sys.stderr
        if run_frontend:
            runFrontEnd()
        self.setUp(title)

    def setUp(self, title: str | None) -> None:
        self.version = self.version_(2)
        if self.version[:2] != "2.":
            raise DebconfError(256, "wrong version: %s" % self.version)
        self.capabilities = self.capb().split()
        if title:
            self.title(title)

    def setCommand(self, command: str) -> None:
        setattr(
            self,
            command,
            lambda *args, **kw: self.command(command, *args, **kw),
        )

    def command(self, command: str, *params: str | int) -> str:
        if command == "version_":
            command = "version"
        command = command.upper()
        self.write.write("{} {}\n".format(command, " ".join(map(str, params))))
        self.write.flush()

        while True:
            try:
                resp = self.read.readline().rstrip("\n")
                break
            except OSError as e:
                if e.errno == errno.EINTR:
                    continue
                else:
                    raise

        if " " in resp:
            status_, data = resp.split(" ", 1)
        else:
            status_, data = resp, ""
        status = int(status_)
        if status == 0:
            return data
        elif status == 1:  # unescaped data
            unescaped = ""
            for chunk in re.split(r"(\\.)", data):
                if chunk.startswith("\\") and len(chunk) == 2:
                    if chunk[1] == "n":
                        unescaped += "\n"
                    else:
                        unescaped += chunk[1]
                else:
                    unescaped += chunk
            return unescaped
        else:
            raise DebconfError(status, data)

    def stop(self) -> None:
        self.write.write("STOP\n")
        self.write.flush()

    def forceInput(self, priority: str, question: str) -> int:
        try:
            self.input(priority, question)
            return 1
        except DebconfError as e:
            if e.args[0] != 30:
                raise
        return 0

    def getBoolean(self, question: str) -> bool:
        result = self.get(question)
        return result == "true"

    def getString(self, question: str) -> str:
        return self.get(question)

    def __enter__(self) -> Debconf:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.stop()


class DebconfCommunicator(Debconf):
    def __init__(
        self, owner: str, title: str | None = None, cloexec: bool = False
    ) -> None:
        args = ["debconf-communicate", "-fnoninteractive", owner]
        self.dccomm: subprocess.Popen[str] | None = subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            close_fds=True,
            universal_newlines=True,
        )
        super().__init__(
            title=title, read=self.dccomm.stdout, write=self.dccomm.stdin
        )
        if cloexec:
            fcntl.fcntl(self.read.fileno(), fcntl.F_SETFD, fcntl.FD_CLOEXEC)
            fcntl.fcntl(self.write.fileno(), fcntl.F_SETFD, fcntl.FD_CLOEXEC)

    def shutdown(self) -> None:
        if self.dccomm is not None:
            assert self.dccomm.stdin is not None
            assert self.dccomm.stdout is not None
            self.dccomm.stdin.close()
            self.dccomm.stdout.close()
            self.dccomm.wait()
            self.dccomm = None

    # Don't rely on this; call .shutdown() explicitly.
    def __del__(self) -> None:
        try:
            self.shutdown()
        except AttributeError:
            pass


if (
    "DEBCONF_USE_CDEBCONF" in os.environ
    and os.environ["DEBCONF_USE_CDEBCONF"] != ""
):
    _frontEndProgram = "/usr/lib/cdebconf/debconf"
else:
    _frontEndProgram = "/usr/share/debconf/frontend"


def runFrontEnd(*, pass_sys_executable: bool = False) -> None:
    if "DEBIAN_HAS_FRONTEND" not in os.environ:
        os.environ["PERL_DL_NONLAZY"] = "1"
        args = [_frontEndProgram]
        if pass_sys_executable:
            args.append(sys.executable)
        args.extend(sys.argv)
        os.execv(_frontEndProgram, args)


if __name__ == "__main__":
    runFrontEnd(pass_sys_executable=True)
    db = Debconf()
    db.forceInput(CRITICAL, "bsdmainutils/calendar_lib_is_not_empty")
    db.go()
    less = db.getBoolean("less/add_mime_handler")
    aptlc = db.getString("apt-listchanges/email-address")
    db.stop()
    print(db.version)
    print(db.capabilities)
    print(less)
    print(aptlc)



# Copyright (c) 2009 Julian Andres Klode <jak@debian.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
"""Progress reporting for text interfaces."""
import io
import os
import signal
import sys
import types
from collections.abc import Callable

import apt_pkg

from apt.progress import base

__all__ = ["AcquireProgress", "CdromProgress", "OpProgress"]


def _(msg: str) -> str:
    """Translate the message, also try apt if translation is missing."""
    res = apt_pkg.gettext(msg)
    if res == msg:
        res = apt_pkg.gettext(msg, "apt")
    return res


class TextProgress:
    """Internal Base class for text progress classes."""

    def __init__(self, outfile: io.TextIOBase | None = None) -> None:
        self._file = outfile or sys.stdout
        self._width = 0

    def _write(self, msg: str, newline: bool = True, maximize: bool = False) -> None:
        """Write the message on the terminal, fill remaining space."""
        self._file.write("\r")
        self._file.write(msg)

        # Fill remaining stuff with whitespace
        if self._width > len(msg):
            self._file.write((self._width - len(msg)) * " ")
        elif maximize:  # Needed for OpProgress.
            self._width = max(self._width, len(msg))
        if newline:
            self._file.write("\n")
        else:
            # self._file.write("\r")
            self._file.flush()


class OpProgress(base.OpProgress, TextProgress):
    """Operation progress reporting.

    This closely resembles OpTextProgress in libapt-pkg.
    """

    def __init__(self, outfile: io.TextIOBase | None = None) -> None:
        TextProgress.__init__(self, outfile)
        base.OpProgress.__init__(self)
        self.old_op = ""

    def update(self, percent: float | None = None) -> None:
        """Called periodically to update the user interface."""
        base.OpProgress.update(self, percent)
        if self.major_change and self.old_op:
            self._write(self.old_op)
        self._write("%s... %i%%\r" % (self.op, self.percent), False, True)
        self.old_op = self.op

    def done(self) -> None:
        """Called once an operation has been completed."""
        base.OpProgress.done(self)
        if self.old_op:
            self._write(_("%c%s... Done") % ("\r", self.old_op), True, True)
        self.old_op = ""


class AcquireProgress(base.AcquireProgress, TextProgress):
    """AcquireProgress for the text interface."""

    def __init__(self, outfile: io.TextIOBase | None = None) -> None:
        TextProgress.__init__(self, outfile)
        base.AcquireProgress.__init__(self)
        self._signal: (
            Callable[[int, types.FrameType | None], None] | int | signal.Handlers | None
        ) = None  # noqa
        self._width = 80
        self._id = 1

    def start(self) -> None:
        """Start an Acquire progress.

        In this case, the function sets up a signal handler for SIGWINCH, i.e.
        window resize signals. And it also sets id to 1.
        """
        base.AcquireProgress.start(self)
        self._signal = signal.signal(signal.SIGWINCH, self._winch)
        # Get the window size.
        self._winch()
        self._id = 1

    def _winch(self, *dummy: object) -> None:
        """Signal handler for window resize signals."""
        if hasattr(self._file, "fileno") and os.isatty(self._file.fileno()):
            import fcntl
            import struct
            import termios

            buf = fcntl.ioctl(self._file, termios.TIOCGWINSZ, 8 * b" ")  # noqa
            dummy, col, dummy, dummy = struct.unpack("hhhh", buf)
            self._width = col - 1  # 1 for the cursor

    def ims_hit(self, item: apt_pkg.AcquireItemDesc) -> None:
        """Called when an item is update (e.g. not modified on the server)."""
        base.AcquireProgress.ims_hit(self, item)
        line = _("Hit ") + item.description
        if item.owner.filesize:
            line += " [%sB]" % apt_pkg.size_to_str(item.owner.filesize)
        self._write(line)

    def fail(self, item: apt_pkg.AcquireItemDesc) -> None:
        """Called when an item is failed."""
        base.AcquireProgress.fail(self, item)
        if item.owner.status == item.owner.STAT_DONE:
            self._write(_("Ign ") + item.description)
        else:
            self._write(_("Err ") + item.description)
            self._write("  %s" % item.owner.error_text)

    def fetch(self, item: apt_pkg.AcquireItemDesc) -> None:
        """Called when some of the item's data is fetched."""
        base.AcquireProgress.fetch(self, item)
        # It's complete already (e.g. Hit)
        if item.owner.complete:
            return
        item.owner.id = self._id
        self._id += 1
        line = _("Get:") + f"{item.owner.id} {item.description}"
        if item.owner.filesize:
            line += " [%sB]" % apt_pkg.size_to_str(item.owner.filesize)

        self._write(line)

    def pulse(self, owner: apt_pkg.Acquire) -> bool:
        """Periodically invoked while the Acquire process is underway.

        Return False if the user asked to cancel the whole Acquire process."""
        base.AcquireProgress.pulse(self, owner)
        # only show progress on a tty to not clutter log files etc
        if hasattr(self._file, "fileno") and not os.isatty(self._file.fileno()):
            return True

        # calculate progress
        percent = ((self.current_bytes + self.current_items) * 100.0) / float(
            self.total_bytes + self.total_items
        )

        shown = False
        tval = "%i%%" % percent
        end = ""
        if self.current_cps:
            eta = int(float(self.total_bytes - self.current_bytes) / self.current_cps)
            end = " {}B/s {}".format(
                apt_pkg.size_to_str(self.current_cps),
                apt_pkg.time_to_str(eta),
            )

        for worker in owner.workers:
            val = ""
            if not worker.current_item:
                if worker.status:
                    val = " [%s]" % worker.status
                    if len(tval) + len(val) + len(end) >= self._width:
                        break
                    tval += val
                    shown = True
                continue
            shown = True

            if worker.current_item.owner.id:
                val += " [%i %s" % (
                    worker.current_item.owner.id,
                    worker.current_item.shortdesc,
                )
            else:
                val += " [%s" % worker.current_item.description
            if worker.current_item.owner.active_subprocess:
                val += " %s" % worker.current_item.owner.active_subprocess

            val += " %sB" % apt_pkg.size_to_str(worker.current_size)

            # Add the total size and percent
            if worker.total_size and not worker.current_item.owner.complete:
                val += "/%sB %i%%" % (
                    apt_pkg.size_to_str(worker.total_size),
                    worker.current_size * 100.0 / worker.total_size,
                )

            val += "]"

            if len(tval) + len(val) + len(end) >= self._width:
                # Display as many items as screen width
                break
            else:
                tval += val

        if not shown:
            tval += _(" [Working]")

        if self.current_cps:
            tval += (self._width - len(end) - len(tval)) * " " + end

        self._write(tval, False)
        return True

    def media_change(self, medium: str, drive: str) -> bool:
        """Prompt the user to change the inserted removable media."""
        base.AcquireProgress.media_change(self, medium, drive)
        self._write(
            _(
                "Media change: please insert the disc labeled\n"
                " '%s'\n"
                "in the drive '%s' and press enter\n"
            )
            % (medium, drive)
        )
        return input() not in ("c", "C")

    def stop(self) -> None:
        """Invoked when the Acquire process stops running."""
        base.AcquireProgress.stop(self)
        # Trick for getting a translation from apt
        self._write(
            (
                _("Fetched %sB in %s (%sB/s)\n")
                % (
                    apt_pkg.size_to_str(self.fetched_bytes),
                    apt_pkg.time_to_str(self.elapsed_time),
                    apt_pkg.size_to_str(self.current_cps),
                )
            ).rstrip("\n")
        )

        # Delete the signal again.
        import signal

        signal.signal(signal.SIGWINCH, self._signal)


class CdromProgress(base.CdromProgress, TextProgress):
    """Text CD-ROM progress."""

    def ask_cdrom_name(self) -> str | None:
        """Ask the user to provide a name for the disc."""
        base.CdromProgress.ask_cdrom_name(self)
        self._write(
            _(
                "Please provide a name for this medium, such as "
                "'Debian 2.1r1 Disk 1'"
            ),
            False,
        )
        try:
            return str(input(":"))
        except KeyboardInterrupt:
            return None

    def update(self, text: str, current: int) -> None:
        """Set the current progress."""
        base.CdromProgress.update(self, text, current)
        if text:
            self._write(text, False)

    def change_cdrom(self) -> bool:
        """Ask the user to change the CD-ROM."""
        base.CdromProgress.change_cdrom(self)
        self._write(_("Please insert an installation medium and press enter"), False)
        try:
            return bool(input() == "")
        except KeyboardInterrupt:
            return False



'''apport hook for openssh-client

(c) 2010 Canonical Ltd.
Author: Chuck Short <chuck.short@canonical.com>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.
'''

from apport.hookutils import (
    attach_conffiles,
    attach_related_packages,
    command_output,
)


def add_info(report, ui):
    response = ui.yesno("The contents of your /etc/ssh/ssh_config file "
                        "may help developers diagnose your bug more "
                        "quickly.  However, it may contain sensitive "
                        "information.  Do you want to include it in your "
                        "bug report?")

    if response == None:  # user cancelled
        raise StopIteration

    elif response:
        attach_conffiles(report, 'openssh-client')

    attach_related_packages(report,
        ['ssh-askpass', 'libpam-ssh', 'keychain', 'ssh-askpass-gnome'])
    report['SSHClientVersion'] = command_output(['/usr/bin/ssh', '-V'])



# vim:set fileencoding=utf-8 et ts=4 sts=4 sw=4:
import os
import time

# check if gtk is available
import gi
gi.require_version('Gtk', '3.0')

# This needs to be after the require_version so we get the correct version of
# Gtk # pylint: disable=wrong-import-position
from gi.repository import Gtk  # noqa

from apt_listchanges.frontends import Frontend  # noqa
from apt_listchanges.ALChacks import _  # noqa


class gtk_frd(Frontend):
    def flush_interface(self):
        while Gtk.events_pending():
            Gtk.main_iteration()

    def cb_close(self, widget):
        Gtk.main_quit()

    # pylint: disable=attribute-defined-outside-init
    def show_interface(self):
        """Instantiate and show the window
        Doesn't happen until we are confident we want it to pop up"""
        if self.interface_showing:
            return
        self.interface_showing = True
        self.builder.set_translation_domain("apt-listchanges")
        mydir = os.path.dirname(__file__)
        uipath = os.path.join(mydir, "apt-listchanges.ui")
        self.builder.add_from_file(uipath)
        self.window_main = self.builder.get_object("window_main")
        if self.hide:
            self.window_main.iconify()
        handlers = {
            "on_button_close_clicked": self.cb_close,
            "on_window_main_destroy": self.cb_close,
        }
        self.progressbar_main = self.builder.get_object("progressbar_main")
        self.button_close = self.builder.get_object("button_close")
        self.label_header = self.builder.get_object("label_header")
        self.label_header_descr = self.builder.get_object("label_header_descr")
        self.label_header_descr.hide()
        self.textview_buf = \
            self.builder.get_object("textview_main").get_buffer()

        self.set_title(_('apt-listchanges: Reading changelogs'))
        self.textview_buf.set_text(
            '\n  ' + _('Reading changelogs. Please wait.'))
        self.builder.connect_signals(handlers)
        self.flush_interface()

    def __init__(self, config, *args):
        super().__init__(config, *args)
        self.hide = config.hide
        self.start_time = time.time()
        self.interface_showing = False
        self.deferred_progress = 0.0
        self.builder = Gtk.Builder()

    def display_output(self, text):
        self.show_interface()
        self.label_header_descr.show()
        self.button_close.set_sensitive(True)
        self.textview_buf.set_text(self._render(text))
        self.window_main.set_urgency_hint(True)
        Gtk.main()

    def update_progress(self, diff=1):
        if not self.interface_showing:
            # If we end up not having any content to display, we don't want a
            # window to flash briefly and then disappear just to briefly
            # display a progress bar, so we wait to show the interface until
            # we've been working for a few seconds and it looks like it's
            # going to be at least a few seconds more.
            if time.time() - self.start_time < 2.5 and \
               self.deferred_progress + diff < self.packages_count / 2:
                self.deferred_progress += diff
                return
            self.show_interface()
            self.progress = self.deferred_progress
            self.deferred_progress = 0
        elif not diff:
            return
        if not hasattr(self, 'progress'):
            self.progress = 0.0
        self.progressbar_main.show()
        self.progress += diff
        self.progressbar_main.set_fraction(self.progress / self.packages_count)
        self.progressbar_main.set_text(
            f"{self.progress*100 / self.packages_count}%")
        self.flush_interface()

    def progress_done(self):
        if not self.interface_showing:
            return
        self.progressbar_main.hide()
        self.flush_interface()

    def confirm(self):
        self.show_interface()
        m = Gtk.MessageDialog(self.window_main,
                              Gtk.DialogFlags.MODAL,
                              Gtk.MessageType.QUESTION,
                              Gtk.ButtonsType.YES_NO)
        m.set_default_response(Gtk.ResponseType.YES)
        m.set_markup("<big><b>%s</b></big>\n\n%s" % (
            _("Continue Installation?"),
            _("Select <i>yes</i> to continue with the installation."
              "\nSelect <i>no</i> to abort the installation.",
              "The 'yes' and 'no' should be translated.")
            ))
        if m.run() == Gtk.ResponseType.NO:
            return False
        return True

    def set_title(self, title):
        self.show_interface()
        self.label_header.set_markup(f'<big><b>{title}</b></big>')
        self.window_main.set_title(title)



#!/usr/bin/python3

# grep the names of all Debian native packages out of Sources files
# Copyright (C) 2007 Stefano Zacchiroli <zack@debian.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

import sys

from debian import deb822


if __name__ == '__main__':
    for fname in sys.argv[1:]:
        with open(fname) as f:
            for stanza in deb822.Sources.iter_paragraphs(f):
                pieces = stanza['version'].split('-')
                if len(pieces) < 2:
                    print(stanza['package'])




'''A multi-producer, multi-consumer queue.'''

import threading
import types
from collections import deque
from heapq import heappush, heappop
from time import monotonic as time
try:
    from _queue import SimpleQueue
except ImportError:
    SimpleQueue = None

__all__ = [
    'Empty',
    'Full',
    'ShutDown',
    'Queue',
    'PriorityQueue',
    'LifoQueue',
    'SimpleQueue',
]


try:
    from _queue import Empty
except ImportError:
    class Empty(Exception):
        'Exception raised by Queue.get(block=0)/get_nowait().'
        pass

class Full(Exception):
    'Exception raised by Queue.put(block=0)/put_nowait().'
    pass


class ShutDown(Exception):
    '''Raised when put/get with shut-down queue.'''


class Queue:
    '''Create a queue object with a given maximum size.

    If maxsize is <= 0, the queue size is infinite.
    '''

    def __init__(self, maxsize=0):
        self.maxsize = maxsize
        self._init(maxsize)

        # mutex must be held whenever the queue is mutating.  All methods
        # that acquire mutex must release it before returning.  mutex
        # is shared between the three conditions, so acquiring and
        # releasing the conditions also acquires and releases mutex.
        self.mutex = threading.Lock()

        # Notify not_empty whenever an item is added to the queue; a
        # thread waiting to get is notified then.
        self.not_empty = threading.Condition(self.mutex)

        # Notify not_full whenever an item is removed from the queue;
        # a thread waiting to put is notified then.
        self.not_full = threading.Condition(self.mutex)

        # Notify all_tasks_done whenever the number of unfinished tasks
        # drops to zero; thread waiting to join() is notified to resume
        self.all_tasks_done = threading.Condition(self.mutex)
        self.unfinished_tasks = 0

        # Queue shutdown state
        self.is_shutdown = False

    def task_done(self):
        '''Indicate that a formerly enqueued task is complete.

        Used by Queue consumer threads.  For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items
        have been processed (meaning that a task_done() call was received
        for every item that had been put() into the queue).

        shutdown(immediate=True) calls task_done() for each remaining item in
        the queue.

        Raises a ValueError if called more times than there were items
        placed in the queue.
        '''
        with self.all_tasks_done:
            unfinished = self.unfinished_tasks - 1
            if unfinished <= 0:
                if unfinished < 0:
                    raise ValueError('task_done() called too many times')
                self.all_tasks_done.notify_all()
            self.unfinished_tasks = unfinished

    def join(self):
        '''Blocks until all items in the Queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer thread calls task_done()
        to indicate the item was retrieved and all work on it is complete.

        When the count of unfinished tasks drops to zero, join() unblocks.
        '''
        with self.all_tasks_done:
            while self.unfinished_tasks:
                self.all_tasks_done.wait()

    def qsize(self):
        '''Return the approximate size of the queue (not reliable!).'''
        with self.mutex:
            return self._qsize()

    def empty(self):
        '''Return True if the queue is empty, False otherwise (not reliable!).

        This method is likely to be removed at some point.  Use qsize() == 0
        as a direct substitute, but be aware that either approach risks a race
        condition where a queue can grow before the result of empty() or
        qsize() can be used.

        To create code that needs to wait for all queued tasks to be
        completed, the preferred technique is to use the join() method.
        '''
        with self.mutex:
            return not self._qsize()

    def full(self):
        '''Return True if the queue is full, False otherwise (not reliable!).

        This method is likely to be removed at some point.  Use qsize() >= n
        as a direct substitute, but be aware that either approach risks a race
        condition where a queue can shrink before the result of full() or
        qsize() can be used.
        '''
        with self.mutex:
            return 0 < self.maxsize <= self._qsize()

    def put(self, item, block=True, timeout=None):
        '''Put an item into the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until a free slot is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Full exception if no free slot was available within that time.
        Otherwise ('block' is false), put an item on the queue if a free slot
        is immediately available, else raise the Full exception ('timeout'
        is ignored in that case).

        Raises ShutDown if the queue has been shut down.
        '''
        with self.not_full:
            if self.is_shutdown:
                raise ShutDown
            if self.maxsize > 0:
                if not block:
                    if self._qsize() >= self.maxsize:
                        raise Full
                elif timeout is None:
                    while self._qsize() >= self.maxsize:
                        self.not_full.wait()
                        if self.is_shutdown:
                            raise ShutDown
                elif timeout < 0:
                    raise ValueError("'timeout' must be a non-negative number")
                else:
                    endtime = time() + timeout
                    while self._qsize() >= self.maxsize:
                        remaining = endtime - time()
                        if remaining <= 0.0:
                            raise Full
                        self.not_full.wait(remaining)
                        if self.is_shutdown:
                            raise ShutDown
            self._put(item)
            self.unfinished_tasks += 1
            self.not_empty.notify()

    def get(self, block=True, timeout=None):
        '''Remove and return an item from the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).

        Raises ShutDown if the queue has been shut down and is empty,
        or if the queue has been shut down immediately.
        '''
        with self.not_empty:
            if self.is_shutdown and not self._qsize():
                raise ShutDown
            if not block:
                if not self._qsize():
                    raise Empty
            elif timeout is None:
                while not self._qsize():
                    self.not_empty.wait()
                    if self.is_shutdown and not self._qsize():
                        raise ShutDown
            elif timeout < 0:
                raise ValueError("'timeout' must be a non-negative number")
            else:
                endtime = time() + timeout
                while not self._qsize():
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise Empty
                    self.not_empty.wait(remaining)
                    if self.is_shutdown and not self._qsize():
                        raise ShutDown
            item = self._get()
            self.not_full.notify()
            return item

    def put_nowait(self, item):
        '''Put an item into the queue without blocking.

        Only enqueue the item if a free slot is immediately available.
        Otherwise raise the Full exception.
        '''
        return self.put(item, block=False)

    def get_nowait(self):
        '''Remove and return an item from the queue without blocking.

        Only get an item if one is immediately available. Otherwise
        raise the Empty exception.
        '''
        return self.get(block=False)

    def shutdown(self, immediate=False):
        '''Shut-down the queue, making queue gets and puts raise ShutDown.

        By default, gets will only raise once the queue is empty. Set
        'immediate' to True to make gets raise immediately instead.

        All blocked callers of put() and get() will be unblocked. If
        'immediate', a task is marked as done for each item remaining in
        the queue, which may unblock callers of join().
        '''
        with self.mutex:
            self.is_shutdown = True
            if immediate:
                while self._qsize():
                    self._get()
                    if self.unfinished_tasks > 0:
                        self.unfinished_tasks -= 1
                # release all blocked threads in `join()`
                self.all_tasks_done.notify_all()
            # All getters need to re-check queue-empty to raise ShutDown
            self.not_empty.notify_all()
            self.not_full.notify_all()

    # Override these methods to implement other queue organizations
    # (e.g. stack or priority queue).
    # These will only be called with appropriate locks held

    # Initialize the queue representation
    def _init(self, maxsize):
        self.queue = deque()

    def _qsize(self):
        return len(self.queue)

    # Put a new item in the queue
    def _put(self, item):
        self.queue.append(item)

    # Get an item from the queue
    def _get(self):
        return self.queue.popleft()

    __class_getitem__ = classmethod(types.GenericAlias)


class PriorityQueue(Queue):
    '''Variant of Queue that retrieves open entries in priority order (lowest first).

    Entries are typically tuples of the form:  (priority number, data).
    '''

    def _init(self, maxsize):
        self.queue = []

    def _qsize(self):
        return len(self.queue)

    def _put(self, item):
        heappush(self.queue, item)

    def _get(self):
        return heappop(self.queue)


class LifoQueue(Queue):
    '''Variant of Queue that retrieves most recently added entries first.'''

    def _init(self, maxsize):
        self.queue = []

    def _qsize(self):
        return len(self.queue)

    def _put(self, item):
        self.queue.append(item)

    def _get(self):
        return self.queue.pop()


class _PySimpleQueue:
    '''Simple, unbounded FIFO queue.

    This pure Python implementation is not reentrant.
    '''
    # Note: while this pure Python version provides fairness
    # (by using a threading.Semaphore which is itself fair, being based
    #  on threading.Condition), fairness is not part of the API contract.
    # This allows the C version to use a different implementation.

    def __init__(self):
        self._queue = deque()
        self._count = threading.Semaphore(0)

    def put(self, item, block=True, timeout=None):
        '''Put the item on the queue.

        The optional 'block' and 'timeout' arguments are ignored, as this method
        never blocks.  They are provided for compatibility with the Queue class.
        '''
        self._queue.append(item)
        self._count.release()

    def get(self, block=True, timeout=None):
        '''Remove and return an item from the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        '''
        if timeout is not None and timeout < 0:
            raise ValueError("'timeout' must be a non-negative number")
        if not self._count.acquire(block, timeout):
            raise Empty
        return self._queue.popleft()

    def put_nowait(self, item):
        '''Put an item into the queue without blocking.

        This is exactly equivalent to `put(item, block=False)` and is only provided
        for compatibility with the Queue class.
        '''
        return self.put(item, block=False)

    def get_nowait(self):
        '''Remove and return an item from the queue without blocking.

        Only get an item if one is immediately available. Otherwise
        raise the Empty exception.
        '''
        return self.get(block=False)

    def empty(self):
        '''Return True if the queue is empty, False otherwise (not reliable!).'''
        return len(self._queue) == 0

    def qsize(self):
        '''Return the approximate size of the queue (not reliable!).'''
        return len(self._queue)

    __class_getitem__ = classmethod(types.GenericAlias)


if SimpleQueue is None:
    SimpleQueue = _PySimpleQueue



try:
    from datetime import datetime
except ImportError:
    datetime = None
import logging
import re
from subprocess import PIPE, Popen
from pickle import dumps

log = logging.getLogger(__name__)
PUBLIC_DIR_RE = re.compile(r'.*?/usr/lib/python(\d(?:.\d+)?)/(site|dist)-packages')


class memoize:
    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args, **kwargs):
        key = dumps((args, kwargs))
        if key not in self.cache:
            self.cache[key] = self.func(*args, **kwargs)
        return self.cache[key]


def execute(command, cwd=None, env=None, log_output=None):
    """Execute external commad.

    :param cdw: currennt working directory
    :param env: environment
    :param log_output:
        * opened log file or path to this file, or
        * None if output should be included in the returned dict, or
        * False if output should be redirectored to stdout/stderr
    """
    args = {'shell': False, 'cwd': cwd, 'env': env}
    close = False
    if log_output is False:
        pass
    elif log_output is None:
        args.update(stdout=PIPE, stderr=PIPE)
    elif log_output:
        if isinstance(log_output, str):
            close = True
            log_output = open(log_output, 'a')
        if datetime:
            log_output.write('\n# command executed on {}'.format(datetime.now().isoformat()))
        log_output.write('\n$ {}\n'.format(repr(command)))
        log_output.flush()
        args.update(stdout=log_output, stderr=log_output)

    log.debug('invoking: %r', command)
    with Popen(command, **args) as process:
        stdout, stderr = process.communicate()
        close and log_output.close()
        return dict(returncode=process.returncode,
                    stdout=stdout and str(stdout, 'utf-8'),
                    stderr=stderr and str(stderr, 'utf-8'))



#!/usr/bin/python3
# vim:set fileencoding=utf-8 et ts=4 sts=4 sw=4:
# This file is shared between postinst and config

import configparser
import os
import shutil
import sys
import debconf

PREFIX_SIZE = len('apt-listchanges/')
DEFAULT_SEEN_DB = '/var/lib/apt/listchanges'
SECTION = 'apt'

previous_version = ''

# See the comment at the top of snapshot.py
snapshots_wanted = False
DO_SNAPSHOTS_OPTION = 'capture_snapshots'
SNAPSHOT_DIR_OPTION = 'snapshot_dir'
DEFAULT_SNAPSHOT_DIR = os.path.join(os.path.dirname(DEFAULT_SEEN_DB),
                                    'listchanges-snapshots')


def _tmpl2Key(name):
    return name[PREFIX_SIZE:].replace('-', '_')


def _debug(*args):
    if 'APT_LISTCHANGES_DEBCONF_DEBUG' in os.environ:
        print(*args, file=sys.stderr)


def _handleString(cfgkey, config, template, db, fromConfig):
    _debug("handleString(", template, cfgkey, fromConfig, ")")
    if fromConfig:
        value = config.get(SECTION, cfgkey)
        if value == 'none':
            value = ''
        db.set(template, value)
    else:
        value = db.getString(template)
        if value == '':
            value = 'none'
        config.set(SECTION, cfgkey, value)


def _handleList(cfgkey, config, template, db, fromConfig):
    _debug("handleList(", template, cfgkey, fromConfig, ")")
    if fromConfig:
        value = config.get(SECTION, cfgkey)
        db.set(template, value.lower())
    else:
        value = db.getString(template)
        config.set(SECTION, cfgkey, value)


def _handleBoolean(cfgkey, config, template, db, fromConfig):
    value = config.getboolean(SECTION, cfgkey, fallback=None)
    _debug("handleBoolean(", template, cfgkey, fromConfig,
           "), old config value:", value)
    if fromConfig:
        db.set(template, str(value).lower())
    else:
        newvalue = db.getBoolean(template)
        if value is None or value != newvalue:
            config.set(SECTION, cfgkey, str(newvalue).lower())


def _handleSeen(cfgkey, config, template, db, fromConfig):
    # The 'save-seen' is very special: a path in config file,
    # but in debconf is stored as boolean...
    value = config.get(SECTION, cfgkey, fallback=None)
    _debug("handleSeen(", template, cfgkey, fromConfig,
           "), old config value:", value)
    if fromConfig:
        db.set(template, str(value and value != 'none').lower())
    elif not db.getBoolean(template):
        value = 'none'
    elif not value or value == 'none':
        value = DEFAULT_SEEN_DB
    if previous_version and previous_version < '4.0' and value.endswith('.db'):
        value = value[:-3]
    config.set(SECTION, cfgkey, value)


IDX_QUESTION_NUMBER = 0
IDX_PRIORITY        = 1
IDX_HANDLER         = 2
IDX_SKIP            = 3
IDX_CHECK           = 4


def _checkFrontend(frontend):
    if frontend == 'none':
        return True

    NAMES['apt-listchanges/confirm'][IDX_SKIP] = frontend == 'mail'
    return False


def _checkWhich(which):
    NAMES['apt-listchanges/no-network'][IDX_SKIP] = which == 'news'
    return False


def _checkEmailAddress(emailAddress):
    NAMES['apt-listchanges/email-format'][IDX_SKIP] = not emailAddress
    return False


NAMES = {'apt-listchanges/frontend'     : [1, debconf.MEDIUM, _handleList,    False, _checkFrontend],
         'apt-listchanges/which'        : [2, debconf.MEDIUM, _handleList,    False, _checkWhich],
         'apt-listchanges/no-network'   : [3, debconf.HIGH,   _handleBoolean, False, None],
         'apt-listchanges/email-address': [4, debconf.LOW,    _handleString,  False, _checkEmailAddress],
         'apt-listchanges/email-format' : [5, debconf.LOW,    _handleList,    False, None],
         'apt-listchanges/confirm'      : [6, debconf.LOW,    _handleBoolean, False, None],
         'apt-listchanges/headers'      : [7, debconf.LOW,    _handleBoolean, False, None],
         'apt-listchanges/reverse'      : [8, debconf.LOW,    _handleBoolean, False, None],
         'apt-listchanges/save-seen'    : [9, debconf.LOW,    _handleSeen,    False, None]
         }


def _updateDebconfFromConfig(config, db):
    _debug("updateDebconfFromConfig()")
    for tmpl, params in NAMES.items():  # don't need to be sorted
        cfgkey = _tmpl2Key(tmpl)
        if config.has_option(SECTION, cfgkey):
            params[IDX_HANDLER](cfgkey, config, tmpl, db, True)
            db.fset(tmpl, 'seen', 'true')


def _communicateWithDebconf(config, db, is_postinst):
    _debug("communicateWithDebconf(", is_postinst, ")")

    for tmpl, params in sorted(NAMES.items(),
                               key=lambda x : x[1][IDX_QUESTION_NUMBER]):
        if params[IDX_SKIP]:
            _debug(f"skipping {tmpl}")
            continue

        if is_postinst:  # store the value
            params[IDX_HANDLER](_tmpl2Key(tmpl), config, tmpl, db, False)
        else:  # ask for the value
            db.forceInput(params[IDX_PRIORITY], tmpl)
            if params[IDX_CHECK]:
                db.go()

        if params[IDX_CHECK]:
            value = db.get(tmpl)
            if params[IDX_CHECK](value):
                _debug(f"finishing on {tmpl}")
                return

    if not is_postinst:
        db.go()


def main(argv):
    global previous_version  # pylint: disable=global-statement

    if len(argv) < 3:
        print("Usage: script postinst|config config_file mainscript_params",
              file=sys.stderr)
        sys.exit(1)

    debconf.runFrontEnd()

    is_postinst = argv[1] == 'postinst'  # otherwise it is config
    config_file = argv[2]
    previous_version = argv[3]
    _debug("apt-listchanges debconf script started(",
           is_postinst, config_file, ")")

    config = configparser.ConfigParser()
    config.read(config_file)

    if not config.has_section(SECTION):
        config.add_section(SECTION)

    try:
        output = os.fdopen(3, "wt")
    except Exception as ex:
        _debug("failed to open file descriptor 3", str(ex))
        output = sys.stdout

    db = debconf.Debconf(write=output)

    if not is_postinst:
        _updateDebconfFromConfig(config, db)

    _communicateWithDebconf(config, db, is_postinst)

    if snapshots_wanted:
        if not config.has_option(SECTION, DO_SNAPSHOTS_OPTION):
            config.set(SECTION, DO_SNAPSHOTS_OPTION, 'auto')
        if not config.has_option(SECTION, SNAPSHOT_DIR_OPTION):
            config.set(SECTION, SNAPSHOT_DIR_OPTION,
                       os.path.join(os.path.dirname(DEFAULT_SEEN_DB),
                                    DEFAULT_SNAPSHOT_DIR))
    elif config.get(SECTION, DO_SNAPSHOTS_OPTION, fallback=None) == 'auto':
        config.remove_option(SECTION, DO_SNAPSHOTS_OPTION)
        snapshot_dir = config.get(SECTION, SNAPSHOT_DIR_OPTION, fallback=None)
        if snapshot_dir:
            shutil.rmtree(snapshot_dir, ignore_errors=True)
        config.remove_option(SECTION, SNAPSHOT_DIR_OPTION)

    if is_postinst:
        with open(config_file + '.new', 'w', encoding='utf-8') as newfile:
            config.write(newfile, space_around_delimiters=False)
            os.fchmod(newfile.fileno(), 0o644)


if __name__ == "__main__":
    main(sys.argv)
    sys.exit(0)



#
# urlutils.py - Simplified urllib handling
#
#   Written by Chris Lawrence <lawrencc@debian.org>
#   (C) 1999-2008 Chris Lawrence
#   Copyright (C) 2008-2023 Sandro Tosi <morph@debian.org>
#
# This program is freely distributable per the following license:
#
#  Permission to use, copy, modify, and distribute this software and its
#  documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appears in all copies and that
#  both that copyright notice and this permission notice appear in
#  supporting documentation.
#
#  I DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL I
#  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
#  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
#  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
#  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
#  SOFTWARE.

import http.client
import urllib.request
import urllib.error
import socket
import shlex
import os
import sys
import webbrowser
import requests

from .exceptions import (
    NoNetwork,
)

from .__init__ import VERSION_NUMBER

UA_STR = 'reportbug/' + VERSION_NUMBER + ' (Debian)'


def urlopen(url, proxies=None, timeout=60, data=None):
    """
    Open an URL and return the content

    This is a helper function for :func:`open_url()`.

    Parameters
    ----------
    url : str
        The URL to retrieve
    proxies : dict
        proxies to use
    timeout : int
        request timeout in seconds
    data
        unused

    Returns
    -------
    str
        Content of the response
    """
    if not proxies:
        proxies = urllib.request.getproxies()

    headers = {'User-Agent': UA_STR,
               'Accept-Encoding': 'gzip;q=1.0, deflate;q=0.9, identity;q=0.5'}

    return requests.get(url, headers=headers, proxies=proxies, timeout=timeout).text


# Global useful URL opener; returns None if the page is absent, otherwise
# like urlopen
def open_url(url, http_proxy=None, timeout=60):
    """
    Open an URL and return the content

    Parameters
    ----------
    url : str
        The URL to retrieve
    http_proxy : str
        HTTP proxy server URL to use for connection.
        By default, use the :func:`urllib.request.getproxies()` settings.
    timeout : int
        connection timeout in seconds

    Returns
    -------
    str
        Content of the response
    """
    # Set timeout to 60 secs (1 min), cfr bug #516449
    # in #572316 we set a user-configurable timeout
    socket.setdefaulttimeout(timeout)

    proxies = urllib.request.getproxies()
    if http_proxy:
        proxies['http'] = http_proxy
        proxies['https'] = http_proxy

    try:
        page = urlopen(url, proxies, timeout)
    except urllib.error.HTTPError as x:
        if x.code in (404, 500, 503):
            return None
        else:
            raise
    except (socket.gaierror, socket.error, urllib.error.URLError):
        raise NoNetwork
    except OSError as data:
        if data and data[0] == 'http error' and data[1] == 404:
            return None
        else:
            raise NoNetwork
    except TypeError:
        print("http_proxy environment variable must be formatted as a valid URI", file=sys.stderr)
        raise NoNetwork
    except http.client.HTTPException as exc:
        exc_name = exc.__class__.__name__
        message = f"Failed to open {url} ({exc_name}: {exc})"
        raise NoNetwork(message)
    return page


def launch_browser(url):
    """
    Launch a web browser to view an URL

    Parameters
    ----------
    url : str
        The URL to view

    Returns
    -------
    None
    """
    if not os.system('command -v xdg-open >/dev/null 2>&1'):
        cmd = 'xdg-open ' + shlex.quote(url)
        os.system(cmd)
        return

    if webbrowser:
        webbrowser.open(url)
        return



# vim:set fileencoding=utf-8 et ts=4 sts=4 sw=4:
#
#   apt-listchanges - Show changelog entries between the installed versions
#                     of a set of packages and the versions contained in
#                     corresponding .deb files
#
#   Copyright (C) 2000-2006  Matt Zimmerman  <mdz@debian.org>
#   Copyright (C) 2006       Pierre Habouzit <madcoder@debian.org>
#   Copyright (C) 2016       Robert Luberda  <robert@debian.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

import os

from apt_listchanges import ALCLog
from apt_listchanges.ALChacks import _


def _parse_apt_bool(value):
    # Based on StringToBool() from apt-pkg/contrib/strutl.cc in apt source
    # and should return same result as StringToBool(value, false)
    return value.lower() in ['1', 'yes', 'true', 'with', 'on', 'enable']


def _parse_apt_int(value):
    # This function should match Configuration::FindI() from apt's
    # apt-pkg/contrib/configuration.cc, except for values like '1something'
    try:
        return int(value)
    except Exception:
        return 0


class AptPipelineError(Exception):
    pass


class AptPipeline:
    def __init__(self, config):
        super().__init__()
        self._config = config

    def read(self):
        ALCLog.debug(_("APT pipeline messages:"))

        with self._open_apt_fd() as fd:
            self._read_version(fd)
            self._read_options(fd)
            debs = self._read_packages(fd)

        ALCLog.debug(_("Packages list:"))
        for d in debs:
            ALCLog.debug(f"\t{d}")
        ALCLog.debug("")

        return debs

    def _open_apt_fd(self):
        if 'APT_HOOK_INFO_FD' not in os.environ:
            raise AptPipelineError(
                _("APT_HOOK_INFO_FD environment variable is not defined\n"
                  "(is Dpkg::Tools::Options::/usr/bin/apt-listchanges::InfoFD "
                  "set to 20?)"))

        try:
            apt_hook_info_fd_val = int(os.environ['APT_HOOK_INFO_FD'])
        except Exception as ex:
            raise AptPipelineError(
                _("Invalid (non-numeric) value of APT_HOOK_INFO_FD"
                  " environment variable")) from ex
        ALCLog.debug(
            _("Will read apt pipeline messages from file descriptor %d") %
            apt_hook_info_fd_val)

        if apt_hook_info_fd_val < 3:
            raise AptPipelineError(
                _("APT_HOOK_INFO_FD environment variable is incorrectly "
                  "defined\n(Dpkg::Tools::Options::/usr/bin/apt-listchanges::"
                  "InfoFD should be greater than 2)."))

        try:
            return os.fdopen(apt_hook_info_fd_val, 'rt')
        except Exception as ex:
            raise AptPipelineError(
                _("Cannot read from file descriptor %(fd)d: %(errmsg)s")
                % {'fd': apt_hook_info_fd_val, 'errmsg': str(ex)}) from ex

    def _read_version(self, fd):
        version = fd.readline().rstrip()
        if version != "VERSION 2":
            raise AptPipelineError(
                _("Wrong or missing VERSION from apt pipeline\n(is Dpkg::"
                  "Tools::Options::/usr/bin/apt-listchanges::Version set to "
                  "2?)"))
        ALCLog.debug(f"\t{version}")

    def _read_options(self, fd):
        while True:
            line = fd.readline().rstrip()
            ALCLog.debug(f"\t{line}")
            if not line:
                return

            if not self._config.ignore_apt_assume and \
               line.startswith('APT::Get::Assume-Yes=') and \
               _parse_apt_bool(line[len('APT::Get::Assume-Yes='):]):
                self._config.confirm = False
                # Set self._config.quiet as well to force non-interactive
                # frontend
                self._config.quiet = max(1, self._config.quiet)
            elif line.startswith('quiet='):
                self._config.quiet = max(_parse_apt_int(line[len('quiet='):]),
                                         self._config.quiet)

    def _read_packages(self, fd):
        filenames = {}
        toconfig = []
        toremove = []
        hasupgrade = False

        for pkgline in fd.readlines():
            pkgline = pkgline.rstrip()
            ALCLog.debug(f"\t{pkgline}")
            if not pkgline:
                break

            (pkgname, oldversion, compare, _newversion, filename) = \
                pkgline.split(None, 4)
            if compare != '<':  # ignore downgrades or re-installations
                continue

            if filename in ('**REMOVE**', '**ERROR**'):
                toremove.append(pkgname)
                continue

            # New installs (oldversion equal to '-') are not ignored to support
            # a case when changelog is moved from one package to a dependent
            # package built from the same source (see p7zip-full 15.09+dfsg-3).
            if oldversion != '-':
                hasupgrade = True

            if filename == '**CONFIGURE**':
                toconfig.append(pkgname)
            else:
                filenames[pkgname] = filename

        # Quit early if no package has been upgraded (e.g. only new installs
        # or removals)
        if not hasupgrade:
            return []

        # Sort by configuration order.  THIS IS IMPORTANT.  Sometimes, a
        # situation exists where package X contains changelog.gz (upstream
        # changelog) and depends on package Y which contains
        # changelog.Debian.gz (Debian changelog).  Until we have a more
        # reliable method for determining whether a package is Debian
        # native, this allows things to work, since Y will always be
        # configured first.

        # apt doesn't explicitly configure everything anymore, so sort
        # the things to be configured first, and then do everything else
        # in alphabetical order.  Also, drop from the list everything
        # that's to be removed.
        for pkg in toremove:
            if pkg in filenames:
                del filenames[pkg]

        ordered_filenames = []
        for pkg in toconfig:
            if pkg in filenames:
                ordered_filenames.append(filenames[pkg])
                del filenames[pkg]

        ordered_filenames.extend(sorted(filenames.values()))
        return ordered_filenames



# urwid user interface for reportbug
#   Written by Chris Lawrence <lawrencc@debian.org>
#   (C) 2006-08 Chris Lawrence
#   Copyright (C) 2008-2023 Sandro Tosi <morph@debian.org>
#
# This program is freely distributable per the following license:
#
#  Permission to use, copy, modify, and distribute this software and its
#  documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appears in all copies and that
#  both that copyright notice and this permission notice appear in
# supporting documentation.
#
#  I DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL I
#  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
#  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
#  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
#  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
#  SOFTWARE.
#
# Portions of this file are licensed under the Lesser GNU Public License
# (LGPL) Version 2.1 or later.  On Debian systems, this license is available
# in /usr/share/common-licenses/LGPL

import locale
import sys
import re
import getpass
from reportbug.exceptions import (
    UINotImportable,
    NoPackage, NoBugs, NoReport, QuertBTSError,
)
from reportbug.urlutils import launch_browser
from .text_ui import (  # noqa: F401
    display_report,
    ewrite,
    spawn_editor,
    system
)
from reportbug import VERSION

try:
    import urwid.raw_display
    import urwid
except ImportError:
    raise UINotImportable('Please install the python3-urwid package to use this interface.')

ISATTY = sys.stdin.isatty()

log_message = ewrite


# Start a urwid session
def initialize_urwid_ui():
    ui = urwid.raw_display.Screen()
    ui.register_palette(palette)
    # Improve responsiveness of UI
    ui.set_input_timeouts(max_wait=0.1)
    return ui


# Empty function to satisfy ui.run_wrapper()
def nullfunc():
    pass


# Widgets ripped mercilessly from urwid examples (dialog.py)
class buttonpush(Exception):
    pass


def button_press(button):
    raise buttonpush(button.exitcode)


class SelectableText(urwid.Edit):
    def valid_char(self, ch):
        return False


class dialog(object):
    def __init__(self, message, body=None, width=None, height=None,
                 title='', long_message=''):
        self.body = body

        self.scrollmode = False
        if not body:
            if long_message:
                box = SelectableText(edit_text=long_message)
                box.set_edit_pos(0)
                self.body = body = urwid.ListBox([box])
                self.scrollmode = True
            else:
                self.body = body = urwid.Filler(urwid.Divider(), 'top')

        if not width:
            width = ('relative', 80)

        if not height:
            height = ('relative', 80)

        self.frame = urwid.Frame(body, focus_part='footer')
        if message:
            self.frame.header = urwid.Pile([urwid.Text(message),
                                            urwid.Divider()])

        w = self.frame
        # pad area around listbox
        w = urwid.Padding(w, ('fixed left', 2), ('fixed right', 2))
        w = urwid.Filler(w, ('fixed top', 1), ('fixed bottom', 1))
        w = urwid.AttrWrap(w, 'body')

        if title:
            w = urwid.Frame(w)
            w.header = urwid.Text(('title', title))

        # "shadow" effect
        w = urwid.Columns([w, ('fixed', 1, urwid.AttrWrap(urwid.Filler(urwid.Text(('border', ' ')), "top"), 'shadow'))])
        w = urwid.Frame(w, footer=urwid.AttrWrap(urwid.Text(('border', ' ')), 'shadow'))
        # outermost border area
        w = urwid.Padding(w, 'center', width)
        w = urwid.Filler(w, 'middle', height)
        w = urwid.AttrWrap(w, 'border')
        self.view = w

    def add_buttons(self, buttons, default=0, vertical=False):
        li = []
        for name, exitcode in buttons:
            if exitcode == '---':
                # Separator is just a text label
                b = urwid.Text(name)
                b = urwid.AttrWrap(b, 'scrolllabel')
            else:
                b = urwid.Button(name, self.button_press)
                b.exitcode = exitcode
                b = urwid.AttrWrap(b, 'selectable', 'focus')
            li.append(b)

        if vertical:
            box = urwid.ListBox(li)
            box.set_focus(default or 0)
            self.buttons = urwid.Frame(urwid.AttrWrap(box, 'selectable'))
            self.frame.footer = urwid.BoxAdapter(self.buttons, min(len(li), 10))
        else:
            self.buttons = urwid.GridFlow(li, 12, 3, 1, 'center')
            self.buttons.set_focus(default or 0)
            self.frame.footer = urwid.Pile([urwid.Divider(), self.buttons],
                                           focus_item=1)

    def button_press(self, button):
        raise buttonpush(button.exitcode)

    def run(self):
        # self.ui.set_mouse_tracking()
        size = self.ui.get_cols_rows()
        try:
            while True:
                canvas = self.view.render(size, focus=True)
                self.ui.draw_screen(size, canvas)
                keys = None
                while not keys:
                    keys = self.ui.get_input()
                for k in keys:
                    if urwid.util.is_mouse_event(k):
                        event, button, col, row = k
                        self.view.mouse_event(size,
                                              event, button, col, row,
                                              focus=True)
                    if k == 'window resize':
                        size = self.ui.get_cols_rows()
                    k = self.view.keypress(size, k)
                    if k:
                        self.unhandled_key(size, k)
        except buttonpush as e:
            return self.on_exit(e.args[0])

    def on_exit(self, exitcode):
        return exitcode

    def unhandled_key(self, size, k):
        if k in ('tab', 'shift tab'):
            focus = self.frame.focus_part
            if focus == 'footer':
                self.frame.set_focus('body')
            else:
                self.frame.set_focus('footer')

        if k in ('up', 'page up', 'down', 'page down'):
            if self.scrollmode:
                self.frame.set_focus('body')
                self.body.keypress(size, k)
            elif k in ('up', 'page up'):
                self.frame.set_focus('body')
            else:
                self.frame.set_focus('footer')

        if k == 'enter':
            # pass enter to the "ok" button
            self.frame.set_focus('footer')
            self.view.keypress(size, k)

    def main(self, ui=None):
        if ui:
            self.ui = ui
        else:
            self.ui = initialize_urwid_ui()
        return self.ui.run_wrapper(self.run)


class displaybox(dialog):
    def show(self, ui=None):
        if ui:
            self.ui = ui
        else:
            self.ui = initialize_urwid_ui()
        size = self.ui.get_cols_rows()
        canvas = self.view.render(size, focus=True)
        self.ui.start()
        self.ui.draw_screen(size, canvas)
        self.ui.stop()


class textentry(dialog):
    def __init__(self, text, width=None, height=None, multiline=False,
                 title='', edit_text=''):
        self.edit = urwid.Edit(edit_text=edit_text, multiline=multiline)
        body = urwid.ListBox([self.edit])
        body = urwid.AttrWrap(body, 'selectable', 'focustext')
        if not multiline:
            body = urwid.Pile([('fixed', 1, body), urwid.Divider()])
            body = urwid.Filler(body)

        dialog.__init__(self, text, body, width, height, title)

        self.frame.set_focus('body')

    def on_exit(self, exitcode):
        return exitcode, self.edit.get_edit_text()


class listdialog(dialog):
    def __init__(self, text, widgets, has_default=False, width=None,
                 height=None, title='', buttonwidth=12):
        li = []
        self.items = []
        for (w, label) in widgets:
            w.others = self.items
            self.items.append(w)
            if label:
                w = urwid.Columns([('fixed', buttonwidth, w),
                                   urwid.Text(label)], 2)
            w = urwid.AttrWrap(w, 'selectable', 'focus')
            li.append(w)

        lb = urwid.ListBox(li)
        lb = urwid.AttrWrap(lb, "selectable")
        dialog.__init__(self, text, height=height, width=width, body=lb,
                        title=title)

        self.frame.set_focus('body')

    def on_exit(self, exitcode):
        """Print the tag of the item selected."""
        if exitcode:
            return exitcode, None

        for i in self.items:
            if hasattr(i, 'get_state') and i.get_state():
                return exitcode, i.get_label()
        return exitcode, None


class checklistdialog(listdialog):
    def on_exit(self, exitcode):
        """
        Mimic dialog(1)'s --checklist exit.
        Put each checked item in double quotes with a trailing space.
        """
        if exitcode:
            return exitcode, []

        li = []
        for i in self.items:
            if i.get_state():
                li.append(i.get_label())
        return exitcode, li


def display_message(message, *args, **kwargs):
    if args:
        message = message % tuple(args)

    if 'title' in kwargs:
        title = kwargs['title']
    else:
        title = ''

    if 'ui' in kwargs:
        ui = kwargs['ui']
    else:
        ui = None

    # Rewrap the message
    chunks = re.split('\n\n+', message)
    chunks = [re.sub(r'\s+', ' ', x).strip() for x in chunks]
    message = '\n\n'.join(chunks).strip()

    box = displaybox('', long_message=message, title=title or VERSION)
    box.show(ui)


def long_message(message, *args, **kwargs):
    if args:
        message = message % tuple(args)

    if 'title' in kwargs:
        title = kwargs['title']
    else:
        title = ''

    if 'ui' in kwargs:
        ui = kwargs['ui']
    else:
        ui = None

    # Rewrap the message
    chunks = re.split('\n+', message)
    chunks = [re.sub(r'\s+', ' ', x).strip() for x in chunks]
    message = '\n\n'.join(chunks).strip()

    box = dialog('', long_message=message, title=title or VERSION)
    box.add_buttons([("OK", 0)])
    box.main(ui)


final_message = long_message
display_failure = long_message


def select_options(msg, ok, help=None, allow_numbers=False, nowrap=False,
                   ui=None, title=None):
    loc = locale.nl_langinfo(locale.CODESET)
    safe_msg = msg.encode(loc, errors='replace').decode(errors='replace')
    box = dialog('', long_message=safe_msg, height=('relative', 80),
                 title=title or VERSION)
    if not help:
        help = {}

    buttons = []
    default = None
    for i, option in enumerate(ok):
        if option.isupper():
            default = i
            option = option.lower()
        buttons.append((help.get(option, option), option))

    box.add_buttons(buttons, default, vertical=True)
    result = box.main(ui)
    return result


def yes_no(msg, yeshelp, nohelp, default=True, nowrap=False, ui=None):
    box = dialog('', long_message=msg, title=VERSION)
    box.add_buttons([('Yes', True), ('No', False)], default=1 - int(default))
    result = box.main(ui)
    return result


def get_string(prompt, options=None, title=None, empty_ok=False, force_prompt=False,
               default='', ui=None):
    if title:
        title = '%s: %s' % (VERSION, title)
    else:
        title = VERSION

    box = textentry(prompt, title=title, edit_text=default)
    box.add_buttons([("OK", 0)])
    code, text = box.main(ui)
    return text or default


def get_multiline(prompt, options=None, title=None, force_prompt=False,
                  ui=None):
    if title:
        title = '%s: %s' % (VERSION, title)
    else:
        title = VERSION

    box = textentry(prompt, multiline=True)
    box.add_buttons([("OK", 0)])
    code, text = box.main(ui)
    li = text.split('\n')
    return li


def get_password(prompt=None):
    return getpass.getpass(prompt)


def _enable_other_checkbox(self, state, other):
    if state and other:
        for widget in self.others:
            if widget.label == other:
                widget.set_state(True)


def menu(par, options, prompt, default=None, title=None, any_ok=False,
         order=None, extras=None, multiple=False, empty_ok=False, ui=None,
         oklabel='Ok', cancellabel='Cancel', quitlabel=None, enable_other={}):
    if not extras:
        extras = []
    else:
        extras = list(extras)

    if multiple:
        if default is None:
            default = []
        elif isinstance(default, str):
            default = [default]
    elif not default:
        default = ''

    if title:
        title = '%s: %s' % (VERSION, title)
    else:
        title = VERSION

    if isinstance(options, dict):
        options = options.copy()
        # Convert to a list
        if order:
            olist = []
            for key in order:
                if key in options:
                    olist.append((key, options[key]))
                    del options[key]

            # Append anything out of order
            options = list(options.items())
            options.sort()
            for option in options:
                olist.append(option)
            options = olist
        else:
            options = list(options.items())
            options.sort()

    opts = []
    for option, desc in options:
        if desc:
            opts.append((option, re.sub(r'\s+', ' ', desc)))
        else:
            opts.append((option, desc))
    options = opts

    if multiple:
        widgets = [(urwid.CheckBox(option, state=(option in default),
                                   on_state_change=_enable_other_checkbox,
                                   user_data=enable_other.get(option, False),
                                   ),
                    desc or '') for (option, desc) in options]
        box = checklistdialog(par, widgets, height=('relative', 80),
                              title=title)
        if quitlabel:
            box.add_buttons([(oklabel, 0), (cancellabel, -1),
                             (quitlabel, -2)])
        else:
            box.add_buttons([(oklabel, 0), (cancellabel, -1)])
        result, chosen = box.main(ui)
        if result < 0:
            # We return None to differentiate a Cancel/Quit from no selection, []
            return None
        return chosen

    # Single menu option only
    def label_button(option, desc):
        return option

    widgets = []
    rlist = []
    for option, desc in options:
        if option == '---':
            # Separator is just a text label
            b = urwid.Text(desc)
            b = urwid.AttrWrap(b, 'scrolllabel')
            desc = ''
        else:
            b = urwid.RadioButton(rlist, label_button(option, desc), state=(option == default))
            b.exitcode = option
            b = urwid.AttrWrap(b, 'selectable', 'focus')
        widgets.append((b, desc))

    box = listdialog(par, widgets, height=('relative', 80),
                     title=title, buttonwidth=12)
    if quitlabel:
        box.add_buttons([(oklabel, 0), (cancellabel, -1), (quitlabel, -2)])
    else:
        box.add_buttons([(oklabel, 0), (cancellabel, -1)])
    focus = 0
    if default:
        for i, opt in enumerate(options):
            if opt[0] == default:
                focus = i  # noqa: F841
                break

    result, chosen = box.main(ui)
    if result < 0:
        return result

    return chosen


# A real file dialog would be nice here
def get_filename(prompt, title=None, force_prompt=False, default=''):
    return get_string(prompt, title=title, force_prompt=force_prompt,
                      default=default)


def select_multiple(par, options, prompt, default=None, title=None,
                    order=None, extras=None, enable_other={}):
    return menu(par, options, prompt, default=default, title=title,
                order=order, extras=extras, multiple=True, empty_ok=False,
                enable_other=enable_other)


# Things that are very UI dependent go here
def show_report(number, bts, mirrors,
                http_proxy, timeout, screen=None, queryonly=False, title='',
                archived='no', mbox_reader_cmd=None):
    from reportbug import debbugs

    ui = screen
    if not ui:
        ui = initialize_urwid_ui()

    sysinfo = debbugs.SYSTEMS[bts]
    display_message('Retrieving report #%d from %s bug tracking system...',
                    number, sysinfo['name'], title=title, ui=ui)

    info = debbugs.get_report(number, timeout, bts, mirrors=mirrors,
                              http_proxy=http_proxy, archived=archived)
    if not info:
        long_message('Bug report #%d not found.', number, title=title, ui=ui)
        return

    options = dict(o='Ok', d='More details (launch browser)',
                   m='Submit more information', q='Quit')
    valid = 'Odmq'

    while 1:
        (buginfo, bodies) = info
        body = bodies[0]

        r = select_options(body, valid, title=buginfo.subject, ui=ui, help=options)
        ui = None
        if not r or (r == 'o'):
            break
        elif r == 'q':
            return -1
        elif r == 'm':
            return buginfo

        launch_browser(debbugs.get_report_url(bts, number, archived))
    return


def handle_bts_query(package, bts, timeout, mirrors=None, http_proxy="",
                     queryonly=False, screen=None, title="", archived='no',
                     source=False, version=None, mbox=False, buglist=None,
                     mbox_reader_cmd=None, latest_first=False):
    from reportbug import debbugs

    sysinfo = debbugs.SYSTEMS[bts]
    root = sysinfo.get('btsroot')
    if not root:
        ewrite("%s bug tracking system has no web URL; bypassing query.\n",
               sysinfo['name'])
        return

    ui = screen
    if not ui:
        ui = initialize_urwid_ui()

    if isinstance(package, str):
        pkgname = package
        if source:
            pkgname += ' (source)'

        display_message('Querying %s bug tracking system for reports on %s',
                        debbugs.SYSTEMS[bts]['name'], pkgname,
                        ui=ui, title=title)
    else:
        display_message('Querying %s bug tracking system for reports %s',
                        debbugs.SYSTEMS[bts]['name'],
                        ' '.join([str(x) for x in package]), ui=ui, title=title)

    result = None
    try:
        (count, sectitle, hierarchy) = debbugs.get_reports(
            package, timeout, bts, mirrors=mirrors, version=version,
            http_proxy=http_proxy, archived=archived, source=source)
    except Exception as e:
        ui.run_wrapper(nullfunc)
        errmsg = 'Unable to connect to %s BTS (error: "%s"); ' % (sysinfo['name'], repr(e))
        raise QuertBTSError(errmsg)

    try:
        if not count:
            ui.run_wrapper(nullfunc)
            if hierarchy is None:
                raise NoPackage
            else:
                raise NoBugs
        else:
            if count > 1:
                sectitle = '%d bug reports found' % (count,)
            else:
                sectitle = '%d bug report found' % (count,)

            buglist = []
            for (t, bugs) in hierarchy:
                buglist.append(('---', t))
                buglist_tmp = {}
                for bug in bugs:
                    # show if the bugs is already resolved
                    done = ''
                    if bug.pending == 'done':
                        done = '  [RESOLVED]'
                    buglist_tmp[bug.bug_num] = bug.subject + done
                # append the sorted list of bugs for this severity
                list(map(buglist.append, [(str(k), buglist_tmp[k]) for k in sorted(buglist_tmp, reverse=latest_first)]))

            p = buglist[1][0]
            # scr.popWindow()
            if queryonly:
                cancellabel = 'Exit'
                quitlabel = None
            else:
                cancellabel = 'New bug'
                quitlabel = 'Quit'

            while True:
                info = int(menu('Select a bug to read (and possibly report more information) or report a new bug:',
                                buglist, '', ui=ui, title=sectitle, default=p,
                                oklabel='Read bug',
                                cancellabel=cancellabel,
                                quitlabel=quitlabel))
                ui = None
                if info < 0:
                    if info == -1:
                        result = None
                    # -2 is the Quit response, triggers the exiting way in main
                    elif info == -2:
                        raise NoReport
                    else:
                        # uniform to return Bugreport instance
                        result = debbugs.get_report(info, timeout)[0]
                    break
                else:
                    res = show_report(info, bts, mirrors, http_proxy,
                                      timeout, queryonly=queryonly)
                    if res:
                        result = res
                        break

    except NoPackage:
        ui.run_wrapper(nullfunc)
        long_message('No record of this package found.', title=title)
        raise NoPackage

    # we didn't find a report; we access Bugreport through debbugs,
    # so to avoid an import of debianbts
    if result and not isinstance(result, debbugs.debianbts.Bugreport):
        raise NoReport

    return result


palette = [
    ('body', 'black', 'light gray', 'standout'),
    ('border', 'black', 'dark blue'),
    ('shadow', 'white', 'black'),
    ('selectable', 'black', 'dark cyan'),
    ('focus', 'white', 'dark blue', 'bold'),
    ('focustext', 'light gray', 'dark blue'),
    ('title', 'dark red', 'light gray'),
    ('scrolllabel', 'white', 'dark cyan'),
]


def initialize():
    return True


def can_input():
    return sys.stdin.isatty()



#! /usr/bin/python3

import os
import re
import sys

_defaults = None
_old_versions = None
_unsupported_versions = None
_supported_versions = ["python%s" % ver.strip() for ver in
                       os.environ.get('DEBPYTHON3_SUPPORTED', '').split(',')
                       if ver.strip()]
#_default_version = "python%s" % os.environ.get('DEBPYTHON3_DEFAULT', '')
#if _default_version == 'python':
#    _default_version = None
_default_version = None


def read_default(name=None):
    global _defaults
    from configparser import ConfigParser, NoOptionError
    if not _defaults:
        if os.path.exists('/usr/share/python3/debian_defaults'):
            config = ConfigParser()
            defaultsfile = open('/usr/share/python3/debian_defaults')
            config.read_file(defaultsfile)
            defaultsfile.close()
            _defaults = config

    if _defaults and name:
        try:
            value = _defaults.get('DEFAULT', name)
        except NoOptionError:
            raise ValueError
        return value
    return None


def version_to_tuple(version):
    return tuple(int(part) for part in version.split('.'))


def parse_versions(vstring):
    if len(vstring.split(',')) > 2:
        raise ValueError('too many arguments provided for X-Python3-Version: min and max only.')
    import operator
    operators = {None: operator.eq, '=': operator.eq,
                 '>=': operator.ge, '<=': operator.le,
                 '<<': operator.lt}
    vinfo = {}
    exact_versions = set()
    version_range = set(supported_versions(version_only=True))
    relop_seen = False
    for field in vstring.split(','):
        field = field.strip()
        if field == 'all':
            continue
        if field in ('current', 'current_ext'):
            continue
        vinfo.setdefault('versions', set())
        ve = re.compile(r'(>=|<=|<<|=)? *(\d\.\d+)$')
        m = ve.match(field)
        try:
            if not m:
                raise ValueError('error parsing Python3-Version attribute')
            op, v = m.group(1), m.group(2)
            vmaj, vmin = v.split('.')
            if int(vmaj) < 3:
                continue
            if op in (None, '='):
                exact_versions.add(v)
            else:
                relop_seen = True
                filtop = operators[op]
                version_range = [av for av in version_range if filtop(
                    version_to_tuple(av), version_to_tuple(v))]
        except Exception:
            raise ValueError('error parsing Python3-Version attribute')
    if 'versions' in vinfo:
        vinfo['versions'] = exact_versions
        if relop_seen:
            vinfo['versions'] = exact_versions.union(version_range)
    return vinfo


def old_versions(version_only=False):
    global _old_versions
    if not _old_versions:
        try:
            value = read_default('old-versions')
            _old_versions = [s.strip() for s in value.split(',')]
        except ValueError:
            _old_versions = []
    if version_only:
        return [v[6:] for v in _old_versions]
    else:
        return _old_versions


def unsupported_versions(version_only=False):
    global _unsupported_versions
    if not _unsupported_versions:
        try:
            value = read_default('unsupported-versions')
            _unsupported_versions = [s.strip() for s in value.split(',')]
        except ValueError:
            _unsupported_versions = []
    if version_only:
        return [v[6:] for v in _unsupported_versions]
    else:
        return _unsupported_versions


def supported_versions(version_only=False):
    global _supported_versions,_default_version
    default_version()
    if not _supported_versions:
        try:
            value = read_default('supported-versions')
            _supported_versions = [s.strip() for s in value.split(',')]
        except ValueError:
            cmd = ['/usr/bin/apt-cache', '--no-all-versions',
                   'show', 'python3-all']
            try:
                import subprocess
                p = subprocess.Popen(cmd, bufsize=1,
                                     shell=False, stdout=subprocess.PIPE)
                fd = p.stdout
            except ImportError:
                fd = os.popen(' '.join(cmd))
            depends = None
            for line in fd:
                if line.startswith('Depends:'):
                    depends = line.split(':', 1)[1].strip().split(',')
            fd.close()
            depends = [re.sub(r'\s*(\S+)[ (]?.*', r'\1', s) for s in depends]
            _supported_versions = depends
    default = _supported_versions.pop(_supported_versions.index(_default_version))
    _supported_versions.sort()
    _supported_versions.append(default)
    if version_only:
        return [v[6:] for v in _supported_versions]
    else:
        return _supported_versions


def default_version(version_only=False):
    global _default_version
    if not _default_version:
        _default_version = os.readlink('/usr/bin/python3')
    # consistency check
    debian_default = read_default('default-version')
    if not _default_version in (debian_default, os.path.join('/usr/bin', debian_default)):
        raise ValueError("the symlink /usr/bin/python3 does not point to the "
                         "python3 default version. It must be reset "
                         "to point to %s" % debian_default)
    _default_version = debian_default
    if version_only:
        return _default_version[6:]
    else:
        return _default_version


def requested_versions(vstring, version_only=False):
    global _default_version
    default_version()
    versions = None
    vinfo = parse_versions(vstring)
    supported = supported_versions(version_only=True)
    if len(vinfo) == 1:
        versions = vinfo['versions'].intersection(supported)
        vl = []
        for version in versions: vl.append(version)
        try:
            default = vl.pop(vl.index(_default_version[6:]))
        except:
            default = ''
        vl.sort()
        if default:
            vl.append(default)
    else:
        raise ValueError('No supported python3 versions in version string')
    if not versions:
        raise ValueError('empty set of versions')
    if version_only:
        return vl
    else:
        return ['python%s' % v for v in vl]


def installed_versions(version_only=False):
    import glob
    supported = supported_versions()
    versions = [os.path.basename(s)
                for s in glob.glob('/usr/bin/python3.[0-9]') + glob.glob('/usr/bin/python3.[0-9][0-9]')
                if os.path.basename(s) in supported]
    versions.sort()
    if version_only:
        return [v[6:] for v in versions]
    else:
        return versions


def minmax_supported_version(minmax, version_only=False):
    supported_versions_list = supported_versions(True)
    version = minmax(version_to_tuple(ver) for ver in supported_versions_list)
    version_str = '%d.%d' % version
    if version_only:
        return version_str
    else:
        return 'python' + version_str


class ControlFileValueError(ValueError):
    pass


class MissingVersionValueError(ValueError):
    pass


def extract_pyversion_attribute(fn, pkg):
    """read the debian/control file, extract the X-Python3-Version
    field."""

    version = None
    sversion = None
    section = None
    with open(fn, encoding='utf-8') as controlfile:
        lines = [line.strip() for line in controlfile]
    for line in lines:
        if line == '' and section != None:
            if pkg == 'Source':
                break
            section = None
        elif line.startswith('Source:'):
            section = 'Source'
        elif line.startswith('Package: ' + pkg):
            section = pkg
        elif line.lower().startswith('x-python3-version:'):
            if section != 'Source':
                raise ValueError('attribute X-Python3-Version not in Source section')
            sversion = line.split(':', 1)[1].strip()
    if section is None:
        raise ControlFileValueError('not a control file')
    if pkg == 'Source':
        if sversion is None:
            raise MissingVersionValueError('no X-Python3-Version in control file')
        return sversion
    return version


'''
def requested_versions_bis(vstring, version_only=False):
    versions = []
    py_supported_short = supported_versions(version_only=True)
    for item in vstring.split(','):
        v=item.split('-')
        if len(v)>1:
            if not v[0]:
                v[0] = py_supported_short[0]
            if not v[1]:
                v[1] = py_supported_short[-1]
            for ver in py_supported_short:
                try:
                    if version_cmp(ver,v[0]) >= 0 \
                           and version_cmp(ver,v[1]) <= 0:
                        versions.append(ver)
                except ValueError:
                    pass
        else:
            if v[0] in py_supported_short:
                versions.append(v[0])
    versions.sort(version_cmp)
    if not versions:
        raise ValueError('empty set of versions')
    if not version_only:
        versions=['python'+i for i in versions]
    return versions
'''


def main():
    from optparse import OptionParser
    usage = '[-v] [-h] [-d|--default] [-s|--supported] [-i|--installed] '
    '[-r|--requested <version string>|<control file>]'
    parser = OptionParser(usage=usage)
    parser.add_option('-d', '--default',
                      help='print the default python3 version',
                      action='store_true', dest='default')
    parser.add_option('-s', '--supported',
                      help='print the supported python3 versions',
                      action='store_true', dest='supported')
    parser.add_option('-r', '--requested',
                      help='print the python3 versions requested by a build; '
                           'the argument is either the name of a control file '
                           'or the value of the X-Python3-Version attribute',
                      action='store_true', dest='requested')
    parser.add_option('-i', '--installed',
                      help='print the installed supported python3 versions',
                      action='store_true', dest='installed')
    parser.add_option('--min-supported',
                      help='print the minimum supported python3 version',
                      action='store_true', dest='minsupported')
    parser.add_option('--max-supported',
                      help='print the maximum supported python3 version',
                      action='store_true', dest='maxsupported')
    parser.add_option('-v', '--version',
                      help='print just the version number(s)',
                      default=False, action='store_true', dest='version_only')
    opts, args = parser.parse_args()
    program = os.path.basename(sys.argv[0])

    if opts.default and len(args) == 0:
        try:
            print(default_version(opts.version_only))
        except ValueError as msg:
            print("%s:" % program, msg)
            sys.exit(1)
    elif opts.supported and len(args) == 0:
        print(' '.join(supported_versions(opts.version_only)))
    elif opts.installed and len(args) == 0:
        print(' '.join(installed_versions(opts.version_only)))
    elif opts.minsupported and len(args) == 0:
        print(minmax_supported_version(min, opts.version_only))
    elif opts.maxsupported and len(args) == 0:
        print(minmax_supported_version(max, opts.version_only))
    elif opts.requested and len(args) <= 1:
        if len(args) == 0:
            versions = 'debian/control'
        else:
            versions = args[0]
        try:
            if os.path.isfile(versions):
                fn = versions
                try:
                    vstring = extract_pyversion_attribute(fn, 'Source')
                    vs = requested_versions(vstring, opts.version_only)
                except ControlFileValueError:
                    sys.stderr.write("%s: not a control file: %s, "
                                     % (program, fn))
                    sys.exit(1)
                except MissingVersionValueError:
                    sys.stderr.write("%s: no X-Python3-Version in control "
                                     "file, using supported versions\n" %
                                     program)
                    vs = supported_versions(opts.version_only)
            else:
                vs = requested_versions(versions, opts.version_only)
            print(' '.join(vs))
        except ValueError as msg:
            sys.stderr.write("%s: %s\n" % (program, msg))
            sys.exit(1)
    else:
        sys.stderr.write("usage: %s %s\n" % (program, usage))
        sys.exit(1)

if __name__ == '__main__':
    main()
